from abc import ABC
from abc import abstractmethod
from typing import List
from naveen.experiment.config.abstract_config import AbstractConfig  # type: ignore # noqa: E501


class Experiment(ABC):

    def __init__(self, config: AbstractConfig):
        self.config = config

    @abstractmethod
    def get_results(self) -> List[dict]:
        pass


class DemoExperiment(Experiment):
    def get_results(self) -> List[dict]:
        if self.config.model == "test/fixtures/models/en.gz":  # type: ignore
            return [{"prod": "az", "hedonic": 8, "utilitatian": 9},
                    {"prod": "aa", "hedonic": 9, "utilitatian": 10}]
        elif self.config.model == "test/fixtures/models/de.gz":  # type: ignore
            return [{"prod": "az", "hedonic": 12, "utilitatian": 13},
                    {"prod": "aa", "hedonic": 14, "utilitatian": 15}]
        raise ValueError("Bad argument in self.config.model")
        return []  # for the linter




class ConfigValidator(object):

    def __init__(self) -> None:
        pass

    @staticmethod
    def is_valid_config_file(config: dict) -> bool:

        required_fields = ["name", "reshaper", "plotter",
                           "experiment_module", "experiment_class"]

        for field in required_fields:
            assert field in config, "Counld not find field {}".format(field)

        assert " " not in config["name"], "Name cant have spaces"
        assert "/" not in config["name"], "Name cant have slashes"

        return True



from dataclasses import make_dataclass
from typing import List
import json
import yaml  # type: ignore # mypy can't find library
from typing import Tuple
from typing import Any

from naveen.experiment.config.config_validator import ConfigValidator


class DynamicExperimentConfigMaker(object):

    def from_yaml(self, config_file: str) -> dict:
        assert config_file.endswith(".yaml") or config_file.endswith(".yml")
        with open(config_file, 'r') as yaml_input:
            yaml_config_as_dict: dict = yaml.safe_load(yaml_input)
            yaml_config_as_dict["filename"] = config_file
        return self._make_config_object(yaml_config_as_dict)  # type: ignore

        # figuring out type anno here is too hard

    def from_json(self, config_file: str):  # type: ignore

        assert config_file.endswith(".json")
        with open(config_file, "r") as inf:
            json_config_as_dict = json.load(inf)
            json_config_as_dict["filename"] = config_file
        return self._make_config_object(json_config_as_dict)  # type: ignore

    def _make_config_object(self, config):  # type: ignore # types confusing
        validator: ConfigValidator = ConfigValidator()
        assert validator.is_valid_config_file(config)

        # https://stackoverflow.com/questions/52534427/dynamically-add-fields-to-dataclass-objects
        fields: List[Tuple[str, Any]] = []
        fields.append(("name", str))
        fields.append(("reshaper", List[str]))
        fields.append(("plotter", List[str]))
        for k, v in config.items():
            if k not in ["name", "reshaper", "plotter"]:
                fields.append((k, type(v)))

        X = make_dataclass('DynamicConfig', fields)

        # https://www.reddit.com/r/learnpython/comments/9h74no/convert_dict_to_dataclass/
        return X(**config)



import logging


def get_logger(name: str = "logger",          # type: ignore
               file: str = "logs/log.txt",
               level: int = logging.NOTSET):  # type: ignore

    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    fh = logging.FileHandler(file)
    fh.setLevel(level)
    # create console handler with a higher log level
    ch = logging.StreamHandler()
    ch.setLevel(level)
    # create formatter and add it to the handlers
    f: str = '%(asctime)s - %(levelname)s - %(message)s'
    formatter = logging.Formatter(f,
                                  datefmt='%Y-%m-%d %H:%M:%S')
    ch.setFormatter(formatter)
    fh.setFormatter(formatter)

    # add the handlers to logger
    if not logger.hasHandlers():
        logger.addHandler(ch)
        logger.addHandler(fh)

    return logger


if __name__ == "__main__":
    logger = get_logger()
    logger.error('This should go to both console and file')



'''
Filter out a url piped in
'''
import sys
import json
from typing import Set


if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('-e',
                        '--e',
                        '-exclude',
                        '--exclude',
                        type=str,
                        dest="excluded",
                        required=True)
    args = parser.parse_args()

    if sys.stdin.isatty():  # no pipe
        print("- You must pipe input to filter")
        import os
        os._exit(0)

    # assume args.excluded is a text file with urls
    with open(args.excluded, "r") as inf:
        excluded_urls: Set[str] = set(o.replace("\n", "") for o in inf)

    for line in sys.stdin:
        line = json.loads(line)
        url: str = line["url"]  # type: ignore
        if url in excluded_urls:
            pass
        else:
            json.dumps(line)



from collections import defaultdict
from copy import deepcopy
import itertools
from typing import List
import yaml  # type: ignore

from naveen.experiment.config.dynamic_experiment_config_maker import DynamicExperimentConfigMaker  # type: ignore # noqa: E501
from naveen.experiment.config.experiment_setting import ExperimentSetting


class ExperimentGroupParser(object):

    def get_experiment_template_kind(self, template: str) -> str:
        if template.endswith("yaml"):
            return "yaml"
        elif template.endswith("json"):
            return "json"
        else:
            msg = "Only json and yaml configs supported, you gave {}"
            raise ValueError(msg.format(template))

    def get_experiment_config(self, experiment_template: str):  # type: ignore
        experiment_config_parser = DynamicExperimentConfigMaker()
        experiment_config_type = self.get_experiment_template_kind(
            experiment_template)
        if experiment_config_type == 'yaml':
            experiment_config = experiment_config_parser.from_yaml(
                experiment_template)
        elif experiment_config_type == 'json':
            experiment_config = experiment_config_parser.from_json(
                experiment_template)
        return experiment_config

    def load_yaml_config(self, group_config_filename: str) -> dict:
        errormessage: str = "Group config must be yaml"
        assert group_config_filename.endswith(
            "yaml") or group_config_filename.endswith("yml"), errormessage

        # config/experiment_groups/demo.yaml
        with open(group_config_filename, 'r') as yaml_input:
            group_config: dict = yaml.safe_load(yaml_input)
            group_config["filename"] = group_config_filename
        return group_config

    def parse_settings(self, group_config: dict) -> List[List]:
        output = defaultdict(list)
        for setting in group_config["settings"]:
            if setting["type"] == "replacement":
                for value in setting["values"]:
                    experiment_setting = ExperimentSetting(field=setting['field'],  # noqa: E501
                                                           value=value)
                    output[setting["field"]].append(experiment_setting)
            if setting['type'] == "range":
                start = setting["start"]
                end = setting["end"]
                step = setting["step"]
                for value in range(start, end, step):
                    experiment_setting = ExperimentSetting(field=setting["field"],  # noqa: E501
                                                           value=value)
                    output[setting["field"]].append(experiment_setting)
            if setting["type"] == "file":  # the values are listed in a path
                # this is helpful for $ find -type f > path_to/options.txt
                path_to: str = setting["path"]
                with open(path_to, "r") as inf:
                    for value in inf:
                        value = value.replace("\n", "")
                        experiment_setting = ExperimentSetting(field=setting["field"],  # noqa: E501
                                                               value=value)
                        output[setting["field"]].append(experiment_setting)

        settings: List[List] = list(output.values())  # list of lists
        return settings

    def __init__(self, group_config_filename: str) -> None:
        self.group_config = self.load_yaml_config(group_config_filename)
        self.experiment_config = self.get_experiment_config(
            self.group_config["experiment_template"])

    def get_group_name(self) -> str:
        return self.group_config["group_name"]

    # returns list of experiment configs
    def parse_experiment_configs(self) -> list:
        settings = self.parse_settings(self.group_config)  # type: ignore
        possible_experiment_configs: List = [
            i for i in itertools.product(*settings)]
        experiment_configs: list = []
        for experiment_config in possible_experiment_configs:
            new_experiment_config = deepcopy(self.experiment_config)
            for setting in experiment_config:
                setattr(new_experiment_config,
                        setting.field, setting.value)
            experiment_configs.append(new_experiment_config)
        return experiment_configs


if __name__ == "__main__":
    parser = ExperimentGroupParser("config/experiment_groups/demo.yaml")
    for c in parser.parse_experiment_configs():
        print(c)



from pathlib import Path
import time
import os
import sys

from naveen.experiment.config.experiment_group_parser import ExperimentGroupParser  # noqa: E501
from naveen.experiment.experiment_runner import ExperimentRunner  # type: ignore # noqa: E501


class ExperimentGroupRunner(object):

    def __init__(self, config_path: str,  # type: ignore
                 group_directory=None) -> None:  # type: ignore
        assert config_path.endswith("yaml") or config_path.endswith(
            "yml"), "Experiment group config must be yaml"
        self.config_path = config_path
        self.group_config = ExperimentGroupParser(
            group_config_filename=self.config_path)
        self.experiment_configs = self.group_config.parse_experiment_configs()
        group_name = self.group_config.get_group_name()
        if group_directory is None:
            self.group_directory = group_name + str(int(time.time()))
        else:
            self.group_directory = group_directory

    def get_path_to_results(self, outputdir: str) -> str:
        path_to_results = os.path.join(outputdir, self.group_directory)
        return path_to_results

    def prepare_path_to_results(self, path_to_results: str) -> None:
        Path(path_to_results).mkdir(parents=True, exist_ok=True)
        # write group config to root directory
        os.system("cp {} {}".format(self.config_path, path_to_results))

    def run(self, outputdir: str, package_name: str) -> None:

        path_to_results = self.get_path_to_results(outputdir)
        self.prepare_path_to_results(path_to_results=path_to_results)

        for ex_no, experiment_config in enumerate(self.experiment_configs):
            path_to_result = os.path.join(outputdir,
                                          self.group_directory,
                                          str(ex_no))
            runner = ExperimentRunner(experiment_config,
                                      package_name=package_name,
                                      output_directory=path_to_result)
            runner.run()


if __name__ == "__main__":
    runner = ExperimentGroupRunner("config/experiment_groups/demo.yaml")
    package_name = "naveen"
    sys.path.append(package_name)
    runner.run(outputdir="tmp", package_name=package_name)
    # rm -rf tmp/demogroup



from bs4 import BeautifulSoup
from bs4.element import ResultSet
import json


class HTMLExtractor(object):

    def __init__(self, html_str: str) -> None:
        self.soup = BeautifulSoup(html_str, 'lxml')

    def select_by_attr(self,
                       attr: str = "class",  # type: ignore
                       html_element: str = "div",  # type: ignore
                       attr_value: str = "bbWrapper") -> ResultSet:
        '''
        e.g. to select all divs with data-role attribute equal to comment
        content do  extractor.select_by_class(attr="data-role",
                                             html_element="div",
                                             attr_value="commentContent")

        to select all p w/ class "A" do
                content do  extractor.select_by_class(attr="class",
                                             html_element="p",
                                             attr_value="A")
        '''
        out = self.soup.find_all(html_element, {attr: attr_value})
        return out


if __name__ == "__main__":
    with open("tmp/index.html", "r") as inf:
        html_doc = inf.read()
        extractor = HTMLExtractor(html_doc)
        messages = extractor.select_by_attr(attr="data-role",
                                            html_element="div",
                                            attr_value="commentContent")
        for i, message in enumerate(messages):
            print(json.dumps({"html": str(message), "i": i}))



import re


class HTMLTagStripper(object):
    '''
    removes elements with a given tag from html string
    See below
    '''

    def __init__(self, html_str: str, tag_to_strip: str):
        self.html_str = html_str
        self.tag_to_strip = tag_to_strip

    def strip(self) -> str:
        '''
        Sample input:
        <p>I\'m in the same spot as you. <abbr title="Licensed Practice Nurse">LPN</abbr>, needing to work from home </p> # noqa: E501
        Sample output:
        <p>I\'m in the same spot as you, needing to work from home </p>
        '''
        regex = "<{}(.|\n)*?</{}>".format(self.tag_to_strip,
                                          self.tag_to_strip)
        return re.sub(regex, "", str(self.html_str))


if __name__ == "__main__":

    with open("test/fixtures/blockquote.html", "r") as inf:
        html = inf.read()

    tag_stripper = HTMLTagStripper(html, "blockquote")

    print(tag_stripper.strip())




from typing import List


class PageMaker(object):

    @staticmethod
    def make_pages(baseurl: str, max_pages: int) -> List[str]:
        output: List[str] = []
        for i in range(1, max_pages + 1):
            output.append(baseurl + str(i))
        return output


if __name__ == "__main__":
    page_maker = PageMaker()
    base: str = "https://allnurses.com/search/?q=EHR&sortby=relevancy&page"
    max_: int = 54
    pages = page_maker.make_pages(base, max_)



import time
import pandas as pd  # type: ignore
import json
import os
from typing import List
from naveen.logger import get_logger
from naveen.experiment.config.abstract_config import AbstractConfig  # type: ignore # noqa: E501


class ExperimentWriter(object):

    def __init__(self,
                 config: AbstractConfig,
                 output_directory: str = None  # type: ignore
                 ) -> None:
        self.config = config
        if output_directory is None:
            output_directory = "results/{}".format(int(time.time()))

            self.output_directory = output_directory
        else:
            self.output_directory = output_directory

        if not os.path.exists(output_directory):
            os.makedirs(output_directory)

        logger = get_logger()

        logger.info(
            "[*] Writing experiment to {}".format(self.output_directory))

    def write_results(self,
                      results: List[dict]
                      ) -> None:
        df = pd.DataFrame(results)

        outpath = os.path.join(self.output_directory,
                               self.config.name + ".csv")
        df.to_csv(outpath, index=False)
        outconfigpath = os.path.join(self.output_directory, "config.json")
        with open(outconfigpath, "w") as of:
            with open(self.config.filename, "r") as inf:
                json.dump(json.load(inf), of)



import importlib
from naveen.experiment.experiment_finisher import ExperimentFinisher
from naveen.experiment.experiment_writer import ExperimentWriter
from naveen.experiment.config.abstract_config import AbstractConfig  # type: ignore  # noqa: E501


class ExperimentRunner(object):

    def __init__(self, config: AbstractConfig,
                 package_name: str = None,  # e.g. src or experiment; name of base package # noqa: E501
                 output_directory: str = None) -> None:  # type: ignore
        self.config = config

        module = importlib.import_module(
            config.experiment_module, package=package_name)
        # figuring out types for next line seems hard and not worth it
        experiment = getattr(module, config.experiment_class)  # type: ignore
        self.experiment = experiment(config)
        self.output_directory = output_directory

    def run(self) -> None:

        if self.output_directory is None:
            writer = ExperimentWriter(self.config)
        else:
            writer = ExperimentWriter(self.config, self.output_directory)
        out = writer.output_directory
        finisher = ExperimentFinisher(self.config,
                                      results_directory=out)
        results = self.experiment.get_results()
        writer.write_results(results)
        finisher.finish_results()
        print("[*] Write results to {}".format(writer.output_directory))
